<?php
/**
 * @file
 * Actually fetch and process wrapped content.
 */

/**
 * Menu callback handler.
 *
 * It's all here!
 *
 * This is what gets called when a 'wrapped' subsection url is requested.
 *
 * @param the node nid that defines the context rules.
 * @param the rest of the args are the menu callback arguments - the
 * contextually requested subpath
 *
 * @return a rendered version of the virtual node
 */
function wrapper_wrap_menu_callback($nid) {
  /////////////////////////////////////////////////////////////////////////////
  // INTERPRET REQUEST

  $args = func_get_args();
  array_shift($args);
  $rel_path = implode("/", $args);

  // When this callback is invoked by way of a 404 handler
  // (neccessary when overlaying the siteroot)
  // The path argument is lost.
  // retrieve it from the internal $_REQUEST  instead
  if (!empty($_REQUEST['destination'])) {
    $rel_path = $_REQUEST['destination'];
  }

  /////////////////////////////////////////////////////////////////////////////
  // Should probably interpret POST and GET also
  // The rel_path is usually a simple string, but if we are being asked for form
  // submission, we need to treat the request as a more complex array
  // - like parse_URL
  $get_data = $_GET;
  unset($get_data['q']);
  $query_string = http_build_query($get_data, '', '&');
  /* see http://www.php.net/manual/en/function.http-build-query.php#102324 */
  if (!empty($query_string)) {
    $rel_request = array(
      'path' => $rel_path,
      'query' => $query_string,
    );
  }
  else {
    $rel_request = $rel_path;
  }


  /////////////////////////////////////////////////////////////////////////////
  // PREPARE the rules.
  if (is_object($nid)) {
    $wrapper = $nid;
    $nid = $wrapper->nid;
  }
  else {
    $wrapper = node_load($nid);
  }
  if (empty($wrapper) || empty($wrapper->source)) {
    // Panic, error.
    drupal_set_message(t("Invalid wrapper [#$nid]. Has it been deleted? May need to rebuild the menu."), 'error');
    return "No content. No wrapper instance.";
  }

  /////////////////////////////////////////////////////////////////////////////
  // CREATE
  // If Create on the fly - let the handler do that entirely.
  // Then return immediately
  if ($wrapper->wrapper_create) {
    unset($wrapper->wrapper_create);
    return wrapper_import_node($wrapper, $rel_request);
    // return is superfluous, it will have finished with a goto (on success)
  }
  // OR
  /////////////////////////////////////////////////////////////////////////////
  // DISPLAY

  $virtual_node = wrapper_fetch_requested_path($wrapper, $rel_request);
  if ($virtual_node) {
    wrapper_emulate_context($virtual_node);
    #dpm($virtual_node);
    return node_view($virtual_node, FALSE, TRUE, TRUE);
  }
  else {
    return t("No content from %rel_request", array('%rel_request' => $rel_request));
  }
}

/**
 * Fetch and parse the wrapped contend, given a pointer to where to get it form.
 *
 * @param the node nid for the wrapper instance that defines the context rules.
 *
 * @param $rel_request the contextually requested subpath. Usually a string, but
 * if it's a GET or POST, then a sructured array.
 *
 * @return a pseudo $node object, NULL on failure.
 */
function wrapper_fetch_requested_path($wrapper, $rel_request) {
  $rel_path = (is_array($rel_request)) ? $rel_request['path'] : $rel_request;

  // Note that we occasionally loose trailing /s from rel_path-
  // Drupal paths won't support that in the initial settings
  // This needs to be worked around later.

  $source_path = rtrim($wrapper->source, '/') . '/' . $rel_path;
  $local_path = rtrim($wrapper->path, '/') . '/' .  $rel_path;

  $strings = array(
    '!source_path' => l($source_path, $source_path),
    '!source' => l($wrapper->source, $wrapper->source),
    '!rel_path' => $rel_path,
    '!wrapper' => l($wrapper->title, 'node/' . $wrapper->nid),
    '!wrapper_path' => $wrapper->path,
    '!local_path' => $local_path,
    '%extraction_method' => $wrapper->extraction_method,
  );

  watchdog('wrapper', "
    Wrapper caught request for '!rel_path', which matches wrapper instance
    '!wrapper'. Returning content of '!source_path' as if it were !local_path
    ",
    $strings, WATCHDOG_INFO
  );

  /////////////////////////////////////////////////////////////////////////////
  // GET THE SOURCE
  // There is a chance that retrieving the target will hit a redirect,
  // in which case $rel_path may be updated by ref. Look out for that.
  // $mime gets filled in by reference
  $mime = '';
  $output = wrapper_retrieve_target($rel_request, $wrapper, $mime);

  if (empty($output)) {
    drupal_set_message(t("Failed to retrieve !rel_path", $strings), "error");
    watchdog("page not found", "Virtual wrapper in wrapper.module failed to retrieve requested !rel_path - expected it to be in !source_path", $strings, WATCHDOG_NOTICE);
    drupal_not_found();
    return NULL;
  }

  // Update the $source_path to reflect any adjustments that a change in $rel_path may have caused.
  $strings['!rel_path'] = $rel_path;
  $source_path = rtrim($wrapper->source, '/') . "/" . $rel_path;
  $strings['!source_path'] = l($source_path, $source_path);
  $strings['!mime'] = $mime;

  // Now either pipe or parse the output depending on the detected content type

  // Bad servers may fail to return a MIME or get it wrong. Good grief.
  // Allow user to set an override for this
  if (empty($mime) || isset($wrapper->wrapper_settings['default_mimes_array']) && in_array($mime, $wrapper->wrapper_settings['default_mimes_array'])) {
    $mime = 'text/html';
  }

  // Passthough method always just sends the file
  // TODO: may need additional headers added
  // If not an HTML page, return the raw content
  if ($wrapper->extraction_method == _WRAPPER_IMPORT_PASSTHRU || $mime != 'text/html') {
    // May need to provide better headers. With remote ones we can probably send the ones we were given
    // For local files, we may need to provide more info
    print($output);
    @watchdog("Wrapper", "
      wrapper.module served !source_path RAW !mime FILE
      , as if it were !wrapper_path!rel_path
      ",
      $strings, WATCHDOG_NOTICE);
    exit;
  }

  // Add an explanation about what's happening - if user is an editor
  if (user_access('access site reports') && ! $wrapper->wrapper_create &&   variable_get('wrapper_show_debug', TRUE)) {
    drupal_set_message(t("
      <strong>Wrapper</strong>:
      Using the wrapper settings !wrapper,
      which mirrors content from !source, extracted using %extraction_method.
      Asked for '<code>!rel_path</code>',
      which will actually be fetched from '!source_path'.
      <br/>
      As it's hosted elsewhere, this content is not directly editable here,
      and the menu structure may not reflect the true state for this location.
      ",
      $strings
    ));
  }

  if (!$output) {
    drupal_set_message(t("File !source_path (!rel_path) is unavailable</a>  ", $strings), "error");
    watchdog("page not found", "Virtual wrapper in wrapper.module failed to retrieve requested $section/$content_path - expected it to be in $source_path", $strings, WATCHDOG_NOTICE);
    drupal_not_found();
    return NULL;
  }

  #dpm(get_defined_vars());

  /////////////////////////////////////////////////////////////////////////////
  // PARSE IT
  // It's not a resource, so parse it and return the tricky page.
  // Note that these pages are not actual nodes yet
  // - they are just content + template made up on the fly,
  // Made up in an object to look like nodes.
  //
  // This includes the links rewriting as needed

  $node = wrapper_absorb($output, $rel_request, $wrapper);

  /////////////////////////////////////////////////////////////////////////////

  if (empty($node->body)) {
    watchdog("Wrapper", "
      Virtual wrapper in wrapper.module failed to format the page retrieved
      from [!source_path] . ($rel_path)
      It did not match the rough content pattern defined by '!wrapper' wrapper.
      Served an unfiltered approximation of what the page should be.
      ", array(
      '!source_path' => l($source_path, $source_path),
      '!wrapper' => l($wrapper->title, 'node/' . $wrapper->nid),
    ), WATCHDOG_NOTICE);
  }

  #dpm(array('Prepared pseudo-node' => $node));

  return $node;
}


/**
 * Go and actually get the target
 *
 * $rel_request is usually a string, but may be an array containing path and
 * query parameters
 *
 * $wrapper is a node that represents an instance of paths to be wrapped. It
 * contains a few top-level properties, such as $wrapper->source, $wrapper-
 * >update_frequency and some sub-properties in a wrapper_settings array, such
 * as $wrapper- >wrapper_settings ['pattern'], which change depending on the
 * $wrapper- >wrapper_method in use.
 */
function wrapper_retrieve_target(&$rel_request, $wrapper, &$mime) {
  // Need to leave a trailing slash on the source URL if one was defined.
  // Otherwise POST to a form that is contained in an index.asp etc fails a bit.

  if (is_array($rel_request)) {
    $rel_path = $rel_request['path'];
  }
  else {
    $rel_path = $rel_request;
  }

  if ($rel_path) {
    $source_path = rtrim($wrapper->source, '/') . "/" . $rel_path;
  }
  else {
    $source_path = $wrapper->source;
  }

  // Local and remote requests have different ways of figuring the filetype,
  // and of handling directories

  if (wrapper_is_local($source_path)) {

    if (is_dir($source_path)) {
      // resolve-select the default document in the case of directory indices
      $directory_indexes = explode(',', $wrapper->wrapper_settings['DirectoryIndex']);
      foreach ($directory_indexes as $directory_index) {
        if (file_exists("$source_path/$directory_index") ) {
          $source_path = "$source_path/$directory_index";
          $rel_path = (($rel_path) ? $rel_path . '/' : "" ) . $directory_index;
        }
      }
    }

    $mime = mime_content_type($source_path);

    // Server-Side Including local files, possibly legacy PHP
    set_include_path(get_include_path() . PATH_SEPARATOR . dirname($source_path) );
    // When running include, this executes the page (which is probably good)
    // but may cause problems when looking for things like the included nav bars...
    if (!empty($wrapper->wrapper_settings['exec_includes'])) {
      // Buffer whatever the page does
      ob_start();
      include($source_path);
      $output = ob_get_contents();
      ob_end_clean();
    }
    else {
      $output = file_get_contents($source_path);
    }
  }
  else {
    // Retrieve REMOTE file
    // Supports passing POST data

    $response_headers = array();
    $may_cache = ($wrapper->update_frequency);
    $original_request = $actual_request = $source_path;
    $output = wrapper_request_remote_file($actual_request, $response_headers, $may_cache);
    // The $actual_request may have been updated if needed
    //- if there was a redirect or query in the process.

    $mime = (!empty($response_headers['Mime'])) ? $response_headers['Mime'] : $response_headers['Content-type'];

    // Check to see if that request was a redirect or the URL was otherwise messed with in the process of retrieving it.

    // Remote hosts may have converted a dir request without a trailing slash
    // into a proper trailing slash
    // (which is more accurate when serving directories or index.html.)
    // I need to know if we are in "/about" or "/about/" when calculating
    // relative links.
    // They are different locations!
    // I have to do this now mostly because the Drupal request handler discards
    // my trailing slashes!
    if ($actual_request != $original_request) {
      // A redirect must have happened

      watchdog('wrapper', "
        When retrieving remote content,
        a remote server redirect must have happened from
        $original_request to $actual_request .
        (often that just means a slash was added at the end)
        This means that relative links are going to be harder to manage
        but I'll try.",
        array(), WATCHDOG_DEBUG
      );

      // TODO Strange things may live under a redirect
      // (like domain switching from www.*)
      // Or a form or htaccess rule putting the browser URL somewhere else.

      // If the retrieval modified the actual_request, that means the $rel_path is out of date.
      // Update it.

      $base = preg_replace('|' . $rel_path . '$|', '', $original_request);
      if ($base && preg_match('|^' . $base . '|', $actual_request)) {
        $new_rel_path = substr($actual_request, strlen($base));
        watchdog('wrapper', "
          Updated rel path [$rel_path] to let the system know that it's actually to be
          [%new_rel_path]",
          array('%new_rel_path' => $new_rel_path), WATCHDOG_INFO
        );
        $rel_path = $new_rel_path;
      }
      // TODO $_GET query strings?
    }
  }
  return $output;
}


/**
 * Create a pseudo node that represents the given source
 *
 * The body of the node will have the parsed output added to it.
 *
 * @param $html_source - as it says, the raw (tidied) HTML source of the page to
 * be emulated
 *
 * @param $rel_request the path relative to the wrapper instance where
 * this page was found and will be served from. $rel_path MAY end with a /, and
 * we need to remember that if relinking.
 *
 * @param a $wrapper definition that will include all the relevant import
 * settings, the rules for how to extract.
 *
 * @return a $node object. Mostly complete, but lacking a nid, as it's not
 * really saved.
 */
function wrapper_absorb($html_source, $rel_request, $wrapper) {
  $rel_path = (is_array($rel_request)) ? $rel_request['path'] : $rel_request;

  if (empty($html_source)) {
    $message = "No HTML source data was found to extract the page from";
    watchdog('wrapper', $message, array(), WATCHDOG_ERROR);
    drupal_set_message($message, 'error');
    return NULL;
  }

  // Create a pseudo-object used for holding the extracted data.
  // It may end up with more than just 'content'.
  // CCK and other attributes may also be added to the $node
  // $node = (object) array();

  // It's probably most convenient to assume that the wrapped page inherits all
  // the properties of the parent item. Though not it's exact 'type'
  $node = clone $wrapper;
  $node->type = isset($wrapper->wrapper_settings['content_type']) ? $wrapper->wrapper_settings['content_type'] : 'wrapped_page';

  // Transfer the wrapper preferences for input format to the node body
  $node->format = $wrapper->wrapper_settings['preferred_filter'];

  // TODO, extend this so as to allow a template node to provide any missing
  // structural background.

  $node->file_data['_original'] = $html_source;
  // source_path is the original url
  $node->source_path = $source_path = rtrim($wrapper->source, '/') . "/" . $rel_path;
  // $node->path is the local drupal url alias.
  // Required to be useful. If not set for the parent,
  // That must mean it's a global (root-level) wrapper.
  // Messing with it in this situation causes local tabs to disappear, ??

  $node->path = ((!empty($wrapper->path)) ? $wrapper->path . '/' : '' ) . $rel_path;

  // This may be needed for the relinking process
  $node->rel_path = $rel_path;

  $node->status = TRUE;

  // Allow other modules to tidy up before we start
  foreach (module_implements('wrapper_pre_parse') as $module) {
    // Want to pass by ref, so can't use call_user_func
    $func_name = $module . '_' . 'wrapper_pre_parse';
    $func_name($html_source, $wrapper, $node);
  }

  // Use the callback defined in the scraper definitions to scrape the content
  // into a pseudo node
  $extraction_methods = wrapper_get_extraction_methods();
  $extraction_method = $extraction_methods[$wrapper->extraction_method];
  $wrapper_absorb_function = $extraction_method['wrapper_absorb_function'];
  $strings = array(
    '%method' => $extraction_method['title'],
    '%func_name' => $wrapper_absorb_function,
  );
  if (function_exists($wrapper_absorb_function)) {
    watchdog('wrapper', "Running {$wrapper_absorb_function} to parse retrieved data for pseudo-node [%path]", array('%path' => $node->path), WATCHDOG_INFO);
    $success = $wrapper_absorb_function($html_source, $wrapper, $node);
    if (!$success) {
      $message = "The %method extraction method failed to return an accurate result.";
      drupal_set_message(t($message, $strings), 'error');
      watchdog('wrapper', $message, $strings, WATCHDOG_INFO);
    }
    if (!empty($node->body)) {
      watchdog('wrapper', "The %method extraction returned some output", $strings, WATCHDOG_INFO);
    }
  }
  else {
    $message = "The %method extraction method doesn't have a useful wrapper_absorb_function. %func_name was not available.";
    drupal_set_message(t($message, $strings), 'error');
    watchdog('wrapper', $message, $strings, WATCHDOG_ERROR);
  }

  // Allow other modules to tweak the node now we are done
  foreach (module_implements('wrapper_post_parse') as $module) {
    // Want to pass by ref, so can't use call_user_func
    $func_name = $module . '_' . 'wrapper_post_parse';
    $func_name($html_source, $wrapper, $node);
  }

  // Although $rel_path may end in a slash, and the relinkers that ran
  // needed to know that, node paths cannot be saved like that, so clean it up
  // to be legal before returning the new object.
  $node->path = rtrim($node->path, '/');

  // Here's a hook for some other modules that change a node before it's
  // used. node_view will do similar fix-ups too, but as this virtual node
  // is sorta like a 'new' node, this nodeapi call is also useful.
  node_invoke_nodeapi($node, 'presave', FALSE, TRUE);

  return $node;
}


/**
 * A callback that will both fetch and absorb the retrieved content as a node,
 * and then save it as new content.
 *
 * Once this is done, it does a goto the new path, to escape from whatever chain
 * of events led us here and to display the result naturally.
 *
 * @arg $wrapper the wrapper node containing import settings
 * @arg $path the rest of the arguments are the $rel_path to import
 */
function wrapper_import_node($wrapper) {
  $args = func_get_args();
  array_shift($args);
  $rel_path = join('/', $args);

  // Need to ensure this is not overwriting an existing node (or merge it?)

  // Ensure the wrapper options are populated with enough context hints
  $wrapper->wrapper_settings['create_menus'] = TRUE;
  $wrapper->wrapper_settings['source_siteroot'] = $wrapper->source;

  $node = wrapper_fetch_requested_path($wrapper, $rel_path);
  // Wrapper absorb functions normally get pseudo-nodes to pretend
  // they are the parent nid. If we want to save a new node, better unset that.
  unset($node->nid);

  if ($node) {
    // I now have the mostly-populated new node.
    // Can save now? Is the result good enough?
    $node = node_submit($node);
    node_save($node);

    // Call after-save hooks that the methods (import_html method)
    // may need to run when making new nodes.
    if ($wrapper->extraction_method == 'import-profile') {
      $wrapper->wrapper_settings += import_html_profile_load($wrapper->wrapper_settings['profile_id']);
      foreach (module_implements('import_html_after_save') as $module) {
        $func = $module . '_import_html_after_save';
        $func($wrapper->wrapper_settings, $node);
      }
    }
    if (user_access('access site reports')) {
      drupal_set_message(t('<strong>Wrapper Import</strong> has SAVED the wrapped version of this page !title as a new %type node [!nid]. This content is now part of the local site and can be edited.', array('!title' => l($node->title, $node->path), '%type' => $node->type, '!nid' => $node->nid)));
    }
  }
  else {
    if (user_access('access site reports')) {
      drupal_set_message(t('<strong>Wrapper Import</strong> has not saved the wrapped version of %rel_path.', array('%rel_path' => $rel_path)));
    }
  }

  drupal_goto($node->path);
}

/**
 * Return the retrieved content without any processing.
 */
function wrapper_passthrough_absorb_function($source, $wrapper, &$node) {
  $node->body = $source;
  return TRUE;
}

/**
 * Run a token-based content extraction on the given source.
 *
 * @param $source is the raw HTML,
 * @param $wrapper is the rules
 * @param $node gets its values filled in by reference
 */
function wrapper_tokens_absorb_function($source, $wrapper, &$node) {
  $pattern = "|"
    . '(' . preg_quote($wrapper->wrapper_settings['token_start']) . ')'
    . '([\s\S]*?)'
    . '(' . preg_quote($wrapper->wrapper_settings['token_finish']) . ')'
    . '|';
  if ( preg_match($pattern, $source, $matches) ) {
    watchdog(__FUNCTION__, 'Got data from supplied source (%source_path) OK.', array('%source_path' => $node->source_path), WATCHDOG_INFO );

    $output = wrapper_redirect_links($matches[2], $node->source_path, $node->path, $wrapper);
    $node->body = $output;

    if ($wrapper->wrapper_settings['innerhtml'] != 'InnerHTML') {
      $node->body = $wrapper->wrapper_settings['token_start'] . $node->body . $wrapper->wrapper_settings['token_finish'];
    }

    return TRUE;
  }
  else {
    watchdog(__FUNCTION__, 'Failed to find any match to the patther %pattern in supplied source (%source_path).', array('%source_path' => $node->source_path, '%pattern' => $pattern), WATCHDOG_INFO );
  }
  return FALSE;
}


/**
 * Run a regular-expression-based content extraction on the given source.
 *
 * @param $source is the raw HTML,
 * @param $wrapper is the rules
 * @param $node gets its values filled in by reference
 */
function wrapper_regexp_absorb_function($source, $wrapper, &$node) {
  $pattern = $wrapper->wrapper_settings['regexp'];
  if ( preg_match($pattern, $source, $matches) ) {
    watchdog(__FUNCTION__, 'Got data from supplied source (%source_path) OK.', array('%source_path' => $node->source_path), WATCHDOG_INFO );
    $output = wrapper_redirect_links($matches[1], $node->source_path, $node->path, $wrapper);
    $node->body = $output;
    return TRUE;
  }
  else {
    watchdog(__FUNCTION__, 'Failed to find any match to the patther %pattern in supplied source (%source_path).', array('%source_path' => $node->source_path, '%pattern' => $pattern), WATCHDOG_INFO );
  }
  return FALSE;
}


/**
 * Run a xpath-based content extraction on the given source.
 *
 * BUG? Some input produces an unwanted number of encoded carriage returns
 * (&#13; or &#xD;) once XML DOM has imported it and returned it. I'm stripping
 * these out in a pre-precess step, but there is probably an encoding directive
 * that can stop this from happening.
 *
 * @param source is the raw HTML,
 * @param $wrapper is the rules
 * @param $node gets its values filled in
 */
function wrapper_xpath_absorb_function($source, $wrapper, &$node) {
  $pattern = $wrapper->wrapper_settings['xpath'];
  if (empty($pattern)) {
    // May be a new or unconfigured node
    watchdog('wrapper', 'No xpath pattern provided', array(), WATCHDOG_WARNING);
    drupal_set_message(t('No xpath pattern provided'), 'warning');
    #dpm(get_defined_vars());
    return NULL;
  }

  $dom = new DOMDocument();
  $success = @$dom->loadHTML($source);

  if (!$success) {
    watchdog('wrapper', 'There was an error parsing the HTML document for xpath extraction', array(), WATCHDOG_ERROR );
    return NULL;
  }

  $xpath = new DOMXpath($dom);

  // find and remove breadcrumbs if the option for that is set. It's more reliable to do this with xpath now than with regexp later.
  if (!empty($wrapper->wrapper_settings['remove_breadcrumb'])) {
    $breadcrumbs = @$xpath->query('//*[@class="breadcrumb"]');
    if (!is_null($breadcrumbs)) {
      foreach ($breadcrumbs as $element) {
        $element->parentNode->removeChild($element);
      }
    }
  }

  // If innerhtml option is set, do not return the element,
  // return the content of the element
  if ($wrapper->wrapper_settings['innerhtml'] == 'InnerHTML') {
    #$pattern .= '/*|text()';
    #$pattern = $pattern . '/*' . '|' . $pattern . '/text()';
    $pattern .= '/node()';
  }

  $all_matches = @$xpath->query($pattern);
  $parsed_items = array();

  if (!is_null($all_matches)) {
    foreach ($all_matches as $element) {
      $parsed_items[] = $dom->saveXML($element);
    }
  }
  if (empty($parsed_items)) {
    watchdog('wrapper', 'XPath pattern %pattern did not match anything within source for %path', array('%pattern' => $pattern, '%path' => $node->path), WATCHDOG_ERROR );
  }

  // Scripts too
  // needed to do this with regexp because xpath did not co-operate with CDATA and nested comments
  // ... as seen in conformant xhtml
  if (!empty($wrapper->wrapper_settings['copy_scripts'])) {
    // If we are copying script tags in ourselves, we do not want to see them
    // inline as well.
    foreach ($parsed_items as $i => $string) {
      $parsed_items[$i] = preg_replace('|<script[\S\s]*?</script>|', '', $parsed_items[$i]);
    }
    $scripts = preg_match_all('|<script[\S\s]*?</script>|', $source, $matches);
    foreach ($matches[0] as $match) {
      $parsed_items[] = $match;
    }
  }
  if (TRUE || $wrapper->wrapper_settings['copy_styles']) {
    $scripts = preg_match_all('|<style[\S\s]*?</style>|', $source, $matches);
    foreach ($matches[0] as $match) {
      drupal_set_html_head($match);
    }
  }

  $output = join("\n", $parsed_items);

  if ($output) {
    watchdog(__FUNCTION__, 'Got data from supplied source (%source_path) OK. Will use it as if it were local at [%path]', array(
      '%source_path' => $node->source_path,
      '%path' => $node->path,
    ), WATCHDOG_INFO );
    $output = wrapper_redirect_links($output, $node->source_path, $node->path, $wrapper);
    $node->body = $output;
    return TRUE;
  }

  #dpm(get_defined_vars());

  return FALSE;
}


/**
 * Run a import_html.module content extraction on the given source.
 *
 * @param $source is the raw HTML,
 * @param $wrapper is the rules
 * @param $node gets its values filled in by reference
 */
function wrapper_import_html_absorb_function($source, $wrapper, &$node) {
  $profile_id = $wrapper->wrapper_settings['profile_id'];
  $wrapper->wrapper_settings += (array) import_html_profile_load($profile_id);

  // import_html expects a path to the source data,
  // but will run if we give it a dummy node also.
  $node->raw_html = $source;
  module_load_include('inc', 'import_html', 'import_html_process');

  import_html_process_html_page($node, $node->path, $wrapper->wrapper_settings);
  return !empty($node->body);
}


function wrapper_import_html_import($rel_path, $wrapper) {
  variable_set('import_html_last_import_timestamp', time());
  $import_html_settings = $wrapper->wrapper_settings;
  $import_html_settings['source_siteroot'] = $wrapper->source;
  $import_html_settings += import_html_profile_defaults();
  return import_html_import_file($rel_path, $import_html_settings);
}


/**
 * Fix up all the links found in the source of an HTML doc.
 *
 * @param $source_path (the documents base_url) includes the wrapper-container
 * URL path as defined in the profile, and the rest of the requested path. Need
 * to chop it back (subtract the $instance['path'] to find the relative path
 *
 * @param $dest_path the localized (probaby root-relative) equivalent of that
 * $source_path. This is the local path that this conent will be served from.
 */
function wrapper_redirect_links($source, $source_path, $dest_path, $wrapper) {

  // Search for inline breadcrumb and removes it
  /* too greedy and dangerous - was wiping out some content? - dman */
  if (!empty($wrapper->wrapper_settings['remove_breadcrumb'])) {
    // remove breadcrumb
    $source = preg_replace('~<div(.*)class="breadcrumb"(.*)</div>~i', '', $source);
    // twice?! wtf? yeah i know it's on my todo list
    $source = preg_replace('~<div(.*)class="breadcrumb"(.*)</div>~i', '', $source);
  }

  $is_root = ($source_path == $wrapper->source);
  // Search for inline major headings and removes them
  if (!empty($wrapper->wrapper_settings['remove_headings']) || (!empty($wrapper->wrapper_settings['remove_root_headings']) && $is_root)) {
    // remove headings
    $source = preg_replace('/<h([1-2]{1})>.*?<\/h\\1>/si', '', $source);
  }
  /* these are not really in the right place, and the regexp needs review for unknonw hazards */

  $strings = array(
    '%source_path' => $source_path,
    '%dest_path' => $dest_path,
  );

  watchdog(__FUNCTION__, '
    Input was from [%source_path], We need to make its links seem as if they
    were under destination path (the local path) [%dest_path].',
    $strings, WATCHDOG_INFO
  );

  // Why is there no PHP func for resolving URLs?

  // To use callbacks, I need to put extra args into global ... :(

  global $_wrapper_new_path; // The localized path of where the current page will be put
  global $_wrapper_instance; // contains ->source (the root of the mirror)

  $_wrapper_instance = $wrapper;
  // Pre-deconstruct the source parts into an array, it's useful later
  $_wrapper_instance->base_parts = parse_url($_wrapper_instance->source);
  // An empty path is OK, but define it to avoid warnings.
  $_wrapper_instance->base_parts += array(
    'path' => '',
    'host' => '',
    'scheme' => '',
  );

  // When linking further items, our drupal path cannot have a trailing slash
  // but the content we put under it will assume it's there.
  // Put it in if needed.
  $_wrapper_instance->path = rtrim($_wrapper_instance->path, '/') . '/'; // ensures a single trailing slash

  // The difference between the wrapper instance root and the current request.
  global $_wrapper_rel_path;
  // The full remote URL of the page being processed
  global $_wrapper_current_path;
  $_wrapper_current_path = $source_path;
  // Broken down version of that.
  global $_wrapper_current_path_parts;
  $_wrapper_current_path_parts = parse_url($_wrapper_current_path);

  $_wrapper_rel_path = $source_path;

  $strings['%wrapper_current_path'] = $_wrapper_current_path;

  if (! $_wrapper_current_path_parts) {
    drupal_set_message(t('
      There was a problem translating %wrapper_current_path into a valid URL.
      URL rewriting will not proceed
    ', $strings), 'error');
    return $source;
  }

  $_wrapper_new_path = $dest_path;

  watchdog('wrapper', "
    Rewriting links within the document from '%wrapper_current_path'
    to its new home at '%dest_path'
    ", $strings, WATCHDOG_INFO
  );
  $output = $source;

  // Replace all src-s
  //$pattern = "|(<[^>]*src=['\"])([^'\"]+?)(['\"][^>]*>)|i";
  // That dies when it hit an asp.net 'viewstate' field with more that 100k in a form element value

  // Try to be correct, while avoiding asp.net stupid fields
  // $pattern = "|(<(img|iframe|script)[^>]*src=['\"])([^'\"]+)(['\"][^>]*>)|i";
  // ... failed.

  // A less intensive, less careful or specific version
  $pattern = "|(src=['\"])([^'\"]+?)(['\"])|i";

  // $match[2] is the URL that needs changing
  // We need to think, so use a callback function
  $output = preg_replace_callback($pattern, 'wrapper_redirect_url_callback', $output);

  if (is_null($output)) {
    watchdog('wrapper', "
      wrapper_redirect_links failed. This is allegedly because of limits on the
      PHP pcre.backtrack_limit value.
      This is known to be triggered when parsing pages containing asp.net VIEWSTATE forms.
      pcre.backtrack_limit=1000000 may fix this
      ", $strings, WATCHDOG_ERROR
    );
    return "Page extraction failed - see the error logs.";
  }

  // Replace all hrefs-s
  //$pattern = "|(<[^>]*href=['\"])([^'\"]+?)(['\"][^>]*>)|i";
  $pattern = "|(href=['\"])([^'\"]+?)(['\"])|i";
  $output = preg_replace_callback($pattern, 'wrapper_redirect_url_callback', $output);

  // and form actions
  // $pattern = "|(<[^>]*action=['\"])([^'\"]+?)(['\"][^>]*>)|i";
  $pattern = "|(action=['\"])([^'\"]+?)(['\"])|i";
  $output = preg_replace_callback($pattern, 'wrapper_redirect_url_callback', $output);

  return $output;
}


/**
 * This is way complex.
 * We have a remote location, possibly a remote subdir, and an actual remote
 * request, which is to be relative to that subdir.
 * We also have a local alias, and the remote rel request is to be rewritten as
 * if it's under the rel path there.
 *
 * Once we've gotten the remote page:
 *
 * Fully-justified external URLs are left alone. - - unless they are actually
 * hard- coded referring to the current site, in which case they are localized
 * temporarily.
 *
 * then the local links are resolved into root-relative ones, fixing all
 * relative links into the actual path.
 *
 * Then the path is compared against the target subdirectory (if any) we are
 * rewriting content for.
 *
 * If it's in, then the subdirectory is removed again to give us the short path
 * that will then be served from under our local alias.
 *
 * Remote pages outside of the mirrored dir get converted into hard, remote URLs
 * back to the source.
 *
 */
function wrapper_redirect_url_callback($matches) {
  // We need to use a bunch of globals to support regexp callbacks.
  // The difference between the wrapper instance root and the current request.
  global $_wrapper_rel_path;
  // The full remote URL of the page being processed
  global $_wrapper_current_path;
  // Broken down version of that.
  global $_wrapper_current_path_parts;
  // The localized path of where the current page will be put
  global $_wrapper_new_path;
  global $_wrapper_instance;

  $link = $matches[2];

  #dpm("Found link to '$link'. May need to shift it from being relative to [source_root]'$_wrapper_current_path' (its home context) to being relative to [wrapper_root]'$_wrapper_new_path' (our new home)" );

  // Rewriting local #anchors (with no other path info) is always unwanted.
  // Leave it alone.
  if (substr($link, 0, 1) == '#') {
    return $matches[0];
  }

  // Calculate/resolve the full URL, based on context.
  // This is MORE complex than what we were given, but more correct, so we don't have to guess anything later.
  $link_parts = parse_url(wrapper_rel2abs($link, $_wrapper_current_path));
  // We'll strip OUT the full path later by replacing the source base with the wrapper base path

  if (! is_array($link_parts) ) {
    drupal_set_message(t("Failed to rewrite found URL : '$link'"), 'error');
    return $matches[0];
  }

  // Full in some blank defaults, even if it's empty, just to avoid warnings.
  $link_parts += array(
    'path' => '',
    'host' => '',
    'scheme' => '',
  );

  // Check if link is a JS action
  // If true leave it alone
  if ($link_parts['scheme'] == 'javascript') {
    return $matches[0];
  }


  if ($link_parts['host'] != $_wrapper_instance->base_parts['host']) {
    // No match to the current context, so it's
    // Absolute, external URL - just return it in one piece
    return $matches[0];
  }

  // Remote link should be resolved and absolute now.

  // We've calculated what the real target path is, but that result may very
  // likely be supposed to fall within our wrapper context, so rename it again.
  //
  // If our local context is different from the remote base, eg if we are renaming
  // a wrapped section of a remote site to be something better on our site, (which is common)
  // then substitute the remote path out for our local new path
  if ($_wrapper_current_path_parts['path'] != '/') {
    $link_parts['path'] = preg_replace('|^' . $_wrapper_instance->base_parts['path'] . '|', $_wrapper_instance->path, $link_parts['path']);
  }

  // Now we know where on the remote location this link really is to
  // But we're not done yet.

  // We want to proxy only content WITHIN the wrapper location,
  // and direct links that fall out of that area probably up-and-over back to the real source.
  // This is no problem if doing a full-site mirror,
  // but may be relevant if only proxying an archive or something

  // Compare the wrapper base path (local) to the new (local) target path
  if (empty($_wrapper_instance->base_parts['path']) || strstr($link_parts['path'], $_wrapper_instance->path)) {
    // if A contains B, it's local to this wrapper context and needs capturing.

    #dpm(__FUNCTION__ . " Path that we will be linking *to* ({$link_parts['path']}) *is* within the realm of this wrapper container, [{$_wrapper_instance->base_parts['path']}], replacing it with {$_wrapper_instance->path}");

    // OK, this is to be mirrored inside the current instance - redirect to me
    // And we prepend our local portal path instead
    unset($link_parts['host']);
    unset($link_parts['port']);
    unset($link_parts['scheme']);
  }
  else {
    #drupal_set_message("path {$link_parts['path']} is NOT inside target [{$_wrapper_instance->source}] [{$_wrapper_instance->base_parts['path']}] - so it's not ours, it should link back offsite (or up and pretend to be local?");
    $link_parts['host'] = $_wrapper_instance->base_parts['host'];
    if (isset($_wrapper_instance->base_parts['port'])) {
      $link_parts['port'] = $_wrapper_instance->base_parts['port'];
    }
    $link_parts['scheme'] = $_wrapper_instance->base_parts['scheme'];
  }

  // When xpath read the link, it may have 'tidied' the ampersands in queries.
  #$link_parts['query'] = html_entity_decode($link_parts['query']);
  #dpm(get_defined_vars());

  $rewritten = wrapper_glue_url($link_parts);

  watchdog('wrapper', "Rewritten <code>%link</code> to <code>%rewritten</code>", array('%link' => $link, '%rewritten' => $rewritten), WATCHDOG_DEBUG);
  return $matches[1] . $rewritten . $matches[3];
}


/**
 * Slightly cleverer parse_url to deal with malformed items.
 * If a link looks broken, try just pegging it to the end of the current
 * context.
 *
 * Some links may appear malformed to parse_url, eg
 * '/cgi-bin/goto/index/2649?http://www.barmixmaster.com/'
 *
 * But we can set them to
 * 'http://webtender.com/cgi-bin/goto/index/2649?http://www.barmixmaster.com/'
 * and continue. (Should technically have been URL-encoded, hence the problem!)
 *
 * If $base is set and some elements are missing from the $link, then transfer
 * parts from the $base into the $link parts
 *
 * @ingroup Utility
 */
function wrapper_parse_url($link, $base = NULL) {
  $link_parts = @parse_url($link);
  if (! is_array($link_parts) ) {
    // Try a messy repair
    $link_parts = $base;
    if (substr($link_parts['path'], 0, 1) == '/') {
      $link_parts['path'] = $link;
    }
    else {
      $link_parts['path'] = dirname($link_parts['path']) . '/' . $link;
    }
    $newlink = wrapper_glue_url($link_parts);
    // try again
    $link_parts = @parse_url($newlink);
    if (! is_array($link_parts) ) {
      return NULL;
    }
    // looks like the patch-up worked
    #dpm(array('repaired link' => $link_parts));
  }

  if (isset($base)) {
    $base_parts = @parse_url($base);
    $base_parts += array(
      'host' => '',
      'scheme' => '',
      'path' => '',
    );
    if (! isset($link_parts['host'])) {
      $link_parts['host'] = $base_parts['host'];
    }
    if (! isset($link_parts['scheme'])) {
      $link_parts['scheme'] = $base_parts['scheme'];
    }
  }

  return $link_parts;
}

/**
 * Resolve a relative link relative to a given base.
 *
 * http://nashruddin.com/PHP_Script_for_Converting_Relative_to_Absolute_URL
 *
 * @param string $rel - relative URL to be resolved
 * @param stringe $base - full URL
 */
function wrapper_rel2abs($rel, $base) {
  /* return if already absolute URL */
  if (parse_url($rel, PHP_URL_SCHEME) != '') {
    return $rel;
  }

  /* queries and anchors */
  if ($rel[0] == '#' || $rel[0] == '?') {
    return $base . $rel;
  }

  /* parse base URL and convert to local variables:
   $scheme, $host, $path */
  $scheme = $host = $path = '';
  extract(parse_url($base));

  /* remove non-directory element from path */
  $path = preg_replace('#/[^/]*$#', '', $path);

  /* destroy path if relative url points to root */
  if ($rel[0] == '/') {
    $path = '';
  }

  /* dirty absolute URL */
  $abs = "$host$path/$rel";

  /* replace '//' or '/./' or '/foo/../' with '/' */
  $re = array('#(/\.?/)#', '#/(?!\.\.)[^/]+/\.\./#');
  for ($n = 1; $n > 0; $abs = preg_replace($re, '/', $abs, -1, $n)) {
  }

  /* absolute URL is ready! */
  return $scheme . '://' . $abs;
}

/**
 * Pus an URL array back into an URL string
 *
 * From http://nz.php.net/parse_url
 *
 * @ingroup Utility
 */
function wrapper_glue_url($parsed) {
  if (!is_array($parsed)) {
    return FALSE;
  }
  $uri = !empty($parsed['scheme']) ? $parsed['scheme'] . ':' . ((strtolower($parsed['scheme']) == 'mailto') ? '' : '//') : '';
  $uri .= !empty($parsed['user']) ? $parsed['user'] . (!empty($parsed['pass']) ? ':' . $parsed['pass'] : '') . '@' : '';
  $uri .= !empty($parsed['host']) ? $parsed['host'] : '';
  $uri .= !empty($parsed['port']) ? ':' . $parsed['port'] : '';
  if (!empty($parsed['path'])) {
    $uri .= (substr($parsed['path'], 0, 1) == '/') ? $parsed['path'] : ('/' . $parsed['path']);
  }
  $uri .= !empty($parsed['query']) ? '?' . $parsed['query'] : '';
  $uri .= !empty($parsed['fragment']) ? '#' . $parsed['fragment'] : '';
  return $uri;
}


/**
 * Fetch remote file.
 *
 * Tries to support a pass-though of HEADERS, COOKIE and POST data
 * to the remote site and support everything that it used to do
 *
 * May cache locally. (Actually it always does, but may or may not delete the
 * cache when done). No caching if a POST is done, but GETs are cacheable
 *
 * @param $url - may become modified by reference if the remote request
 * triggered a redirect or two, including if '/path' became '/path/'
 * It's usually a string, but may be an array containing form data also
 *
 * @param headers array gets passed back by reference
 * @return the basic content of the file
 *
 */
function wrapper_request_remote_file(&$url, &$response_headers, $may_cache = FALSE) {
  global $http_response_header;
  unset($http_response_header);
  $cachepath = wrapper_cache_filepath($url);

  // Immediately return the cache version if available.
  if ($may_cache && is_file($cachepath) && ! $_SERVER['REQUEST_METHOD'] == 'GET') {
    $output = file_get_contents($cachepath);
    $response_headers['Mime'] = mime_content_type($cachepath);
    drupal_set_message(t("Returning cached GET reponse from %cachepath", array('%cachepath' => $cachepath)));
    return $output;
  }

  // Fetching remote file
  $url_parts = parse_url($url);

  $strings = array(
    '!url' => l($url, $url),
    '!cachepath' => l($cachepath, file_create_url($cachepath)),
    '!post_keys' => join(', ', array_keys($_REQUEST)),
  );

  if ($_SERVER['REQUEST_METHOD'] == 'POST') {
    watchdog('wrapper', "Wrapping a POST form submission. Relaying post data to '!url' as best we can. Submited keys include !post_keys", $strings, WATCHDOG_INFO);
    // Cannot POST to GET URLs.
    // If for some reason we think that is what we are doing, (bad rewrite logic elsewhere)
    // then remove those GET values.
    unset($url_parts['query']);
    $url = wrapper_glue_url($url_parts);
  }

  // Set up context headers.
  // So forms can be wrapped and sessions maintained
  // Works best with cURL available! ??

  $http_opts = array();
  // Allow SELECTED session data though to the remote script
  $http_opts['http']['user_agent'] = $_SERVER['HTTP_USER_AGENT'];
  if (isset($_SERVER['HTTP_COOKIE'])) {
    $http_opts['http']['cookie'] = $_SERVER['HTTP_COOKIE'];
    $http_opts['http']['header']['Cookie'] = "Cookie: {$_SERVER['HTTP_COOKIE']}";
  }
  $http_opts['http']['header']['Referer'] = 'Referer: ' . $url;
  $http_opts['http']['method'] = $_SERVER['REQUEST_METHOD'];

  // TODO: Expose this?
  $http_opts['http']['timeout'] = 5;

  // Set up the proxy if one is specified in the administration page.
  $proxy = variable_get('wrapper_use_proxy', '');
  if (!empty($proxy)) {
    $http_opts['http']['proxy'] = $proxy;
    $http_opts['http']['request_fulluri'] = true;
  }

  #dpm($_SERVER);
  #dpm($_REQUEST);

  if ($_SERVER['REQUEST_METHOD'] == 'POST') {
    // Relay post data - THIS WAS TRICKY
    $data = "";
    $boundary = "#####" . substr(md5(rand(0, 32000)), 0, 10) . '#####';
    $http_opts['http']['header']['Content-Type'] = 'Content-Type: multipart/form-data; boundary=' . $boundary;
    // Collect Postdata, re-encode it
    // If the post data was supplied in arrays, PHP would have already structured it for us
    // but the raw http request we are constructing needs it flat again.
    // The is evident when sending to Drupal #tree forms, but is possible anywhere.
    // convert a structured $_REQUEST to a flat labelled one.
    $flat_key_vals = wrapper_flatten_key_vals($_REQUEST);

    foreach ($flat_key_vals as $key => $val) {
      //foreach ($_REQUEST as $key => $val) {
      if ($key != "q") {
        // the extra two '--' have special protocol meaning
        $data .= '--' . $boundary . "\n";
        $data .= "Content-Disposition: form-data; name=\"{$key}\"\n\n{$val}\n";
      }
    }
    $data .= '--' . $boundary . "\n";

    $http_opts['http']['content'] = $data;
  }
  $http_opts['http']['header']['Content-length'] = "Content-length : " . strlen($data);

  // Header as an array is only supported PHP 5.10+
  // http://www.php.net/manual/en/context.http.php
  // IF PHP < 5.10, need to glue the headers together
  $new_header = '';
  foreach ($http_opts['http']['header'] as $key => $val) {
    $new_header .= "$val\r\n";
  }
  $http_opts['http']['header'] = $new_header;
  // - that wasn't needed for newer PHP installs

  $context = stream_context_create($http_opts);

  // It's possible that Arguments are passed in at our end.
  // and we need to know what they are and retain them.
  // Especially if wrapping outside sources. Eg other CMS with ?page=123
  if ($_SERVER['REQUEST_METHOD'] != 'POST') {
    // Don't GET when POSTing
    $get_data = $_GET;
    unset($get_data['q']);
    $query_string = http_build_query($get_data, '', '&');
    // see http://www.php.net/manual/en/function.http-build-query.php#102324
    if (!empty($query_string)) {
      $url .= '?' . $query_string;
      $url_parts['query'] = $query_string;
      $strings['!url'] = $url;
    }
  }

  watchdog('wrapper', "Remotely requesting '!url' ", $strings, WATCHDOG_INFO);

  $output = file_get_contents($url, FALSE, $context );

  #dpm($http_opts);
  #dpm(get_defined_vars());

  // BUG
  // follow_location option doesn't work right
  // unless curl is also working?
  // So if the target returns a Location 301 redirect,
  // Older PHP systems can't handle it and think it's an error..
  // See below for a work-around.

  // Figure out what happened, what the MIME was
  if ($response_headers = wrapper_get_response_headers($http_response_header) ) {

    // If the remote site told us the Location was something different from
    // what we actually asked for, that was a redirect that we may have to care
    // about this when resolving relative links.
    if (isset($response_headers['Location']) && $response_headers['Location'] != $url) {

      // The response (from some servers anyway) overly URL-encodes the response.
      // It's not incorrect, but it's ugly.
      // Substituting %2d for '-' helps nobody.
      // Tidy it for debugging, and comparison
      $actual_request_parts = wrapper_parse_url($response_headers['Location'], $url);
      $actual_request_parts['path'] = urldecode($actual_request_parts['path']);

      // Sometimes a Location redirect is relative, sometimes remote, sometimes absolute but local. Sigh.
      if ($actual_request_parts['host'] == $url_parts['host']) {
        $actual_request_parts['host'] = $_SERVER['HTTP_HOST'];
      }

      $response_location = wrapper_glue_url($actual_request_parts);

      $strings['!location'] = l($response_location, $response_location);
      watchdog('wrapper', "It seems our request for '!url' became a redirect to '!location'. Making a note of that", $strings, WATCHDOG_INFO);

      // IF there was a redirect AND the redirect wasn't just a trailing slash
      // OR there was a redirect that resulted in a different GET query
      // ACTUALLY DO the redirect so the browser URL doesn't get out of sync.
      if (rtrim($actual_request_parts['path'], '/') != rtrim($url_parts['path'], '/')
           || ($actual_request_parts['query'] != $url_parts['path'] )
        ) {
        // Fully emulate what happened.
        watchdog('wrapper', "As the URL changed to !location (not !url), I'm telling the browser to redirect.", $strings, WATCHDOG_INFO);
        drupal_set_header("Location:" . $response_location);
        exit();
      }
    }

    $mime_full = $response_headers['Content-Type'];
    $mime = array_shift(split(';', $response_headers['Content-Type']));
    $response_headers['Mime'] = $mime;

    // If the remote site tried to set a cookie, pass that back to us.
    // With luck, we'll be allowed to proxy it.
    if (isset($response_headers['Set-Cookie'])) {
      header("Set-Cookie : {$response_headers['Set-Cookie']}");
    }

  }
  else {
    watchdog("Wrapper", "No Response headers from '!url'. This should not happen", $strings, 'error');
  }

  // May cache the passthrough content for next time
  if ($may_cache) {
    if ( ! is_dir(dirname($cachepath)) ) {
      mkdirs(dirname($cachepath));
    }
    file_put_contents($cachepath, $output);
    watchdog("wrapper", "Saved a cache of !url in !cachepath", $strings, WATCHDOG_INFO);
  }

  return $output;
}

/**
 * Convert a nested array into flattened form keys.
 *
 * eg
 *   array('submitted' => array('name' => 'joe', 'role' => 'boss'));
 * becomes
 *   array('submitted[name]' => 'joe', 'submitted[role]' => 'boss');
 *
 * Used to prepare http posts.
 *
 * @param array
 */
function wrapper_flatten_key_vals($in, $prefix_list = array()) {
  $results = array();
  foreach ($in as $key => $val) {
    $prefix_list_copy = $prefix_list;
    $prefix_list_copy[] = $key;
    if (is_array($val)) {
      $results += wrapper_flatten_key_vals($val, $prefix_list_copy);
    }
    else {
      $flat_key = array_shift($prefix_list_copy);
      foreach ($prefix_list_copy as $prefix) {
        $flat_key .= "[$prefix]";
      }
      $results[$flat_key] = $val;
    }
  }
  return $results;
}

/**
 * Some work-around to retrospectively get request headers. Keyed this time, as
 * the normal results are not keyed.
 *
 * @ingroup Utility
 */
function wrapper_get_response_headers($http_response_header) {
  foreach ((array) $http_response_header as $wrapper_line) {
    $pair = split(':', $wrapper_line, 2);
    if (isset($pair[1])) {
      // Not all HTTP lines are splitable
      $response[$pair[0]] = trim($pair[1]);
    }
  }
  return $response;
}


/**
 * Returns a safe filepath to save a given URL into.
 *
 * If wrapper caching is
 * enabled, where should the cache files be saved
 *
 *
 * Ensure that directory listings are saved as $path/index.htm
 */
function wrapper_cache_filepath($path) {

  if (preg_match('|/[^\./]+$|', $path)) {
    $path .= '/';
  }
  if (preg_match('|/$|', $path)) {
    $path .= 'index.htm';
  }

  // Collect all argumentative requests in a folder named after the file.
  $path = preg_replace("|\?|", "_/", $path);

  $path = preg_replace("|#.*|", "", $path);
  $path = preg_replace("|http://|", "", $path);
  $path = preg_replace("|[^A-Za-z0-9_\-~\./]+|", "_", $path);
  return file_directory_path() . '/' . variable_get('wrapper_cache_path', 'wrapper') . '/' . $path;
}


/**
 * If this func is unavailable, add it.
 */
if (! function_exists('mkdirs')) {
  /**
   * Recursive directory creation.
   * Very common task, so it may have been declared by other libraries.
   *
   * @ingroup Utility
   */
  function mkdirs($strpath, $mode = 0777) {
    //creates directory tree recursively
    return is_dir($strpath) || ( mkdirs(dirname($strpath), $mode) and mkdir($strpath, $mode) and @chmod($strpath, 0777) );
  }
}


/**
 * Another utility, strange there's no reliable MIME library around.'
 *
 * @ingroup Utility
 */
if ( ! function_exists('mime_content_type') ) {
  function mime_content_type( $f ) {
    return trim(exec('file -bi ' . escapeshellarg($f) ));
  }
}
